'use strict';

const storj = require('storj-lib');
const mongoose = require('mongoose');
const SchemaOptions = require('../options');
const crypto = require('crypto');
const errors = require('storj-service-error-types');
/**
 * Represents a storage bucket
 * @constructor
 */
var Bucket = new mongoose.Schema({
  storage: {
    type: Number,
    default: 0
  },
  transfer: {
    type: Number,
    default: 0
  },
  status: {
    type: String,
    enum: ['Active', 'Inactive'],
    default: 'Active'
  },
  pubkeys: [{
    type: String,
    ref: 'PublicKey'
  }],
  user: {
    type: String,
    ref: 'User'
  },
  name: {
    type: String,
    default: function() {
      return 'Bucket-' + crypto.randomBytes(3).toString('hex');
    }
  },
  created: {
    type: Date,
    default: Date.now
  },
  publicPermissions: {
    type: [{
      type: String,
      enum: ['PUSH', 'PULL'],
    }],
    default: []
  },
  encryptionKey: {
    type: String,
    default: ''
  },
});

Bucket.plugin(SchemaOptions);

Bucket.index({ user: 1 });
Bucket.index({ created: 1 });
Bucket.index({ user: 1, name: 1 }, { unique: true });

Bucket.set('toObject', {
  transform: function(doc, ret) {
    delete ret.__v;
    delete ret._id;

    ret.id = doc._id;
  }
});

/**
 * Creates a Bucket
 * @param {storage.models.User} user
 * @param {Object} data
 * @param {Function} callback
 */
Bucket.statics.create = function(user, data, callback) {
  let Bucket = this;

  let bucket = new Bucket({
    status: 'Active',
    pubkeys: data.pubkeys,
    user: user._id
  });

  if (data.name) {
    bucket.name = data.name;
  }

  // XXX THIS WILL BE DEPRECATED IN THE NEXT MAJOR RELEASE
  // XXX Make sure to update any code that depends on this being
  // XXX based on the name as soon as possible. Once this is
  // XXX deprecated the id will be a unique ObjectId
  bucket._id = mongoose.Types.ObjectId(
    storj.utils.calculateBucketId(user._id, bucket.name)
  );

  bucket.save(function(err) {
    if (err) {
      if (err.code === 11000) {
        return callback(
          new errors.ConflictError('Name already used by another bucket')
        );
      }

      return callback(new errors.InternalError(err.message));
    }

    Bucket.findOne({ _id: bucket._id }, function(err, bucket) {
      if (err) {
        return callback(new errors.InternalError(err.message));
      }

      if (!bucket) {
        return callback(
          new errors.InternalError('Failed to load created bucket')
        );
      }

      callback(null, bucket);
    });
  });
};

module.exports = function(connection) {
  return connection.model('Bucket', Bucket);
};
